@@ -1,4 +1,3 @@
-#!/usr/bin/python
 # encoding=UTF-8
 # Copyright © 2008, 2009, 2010 Jakub Wilk <jwilk@jwilk.net>
 #
@@ -19,17 +18,17 @@ import inspect
 import locale
 import os.path
 import shutil
+import struct
 import sys
 import tempfile
 import threading
 
 import djvu.decode
 
-from . import cuneiform
+from . import engines
 from . import errors
 from . import hocr
 from . import ipc
-from . import tesseract
 from . import utils
 from . import version
 
@@ -106,7 +105,6 @@ class InPlaceSaver(Saver):
         djvu_path = os.path.abspath(djvu_path)
         djvused = ipc.Subprocess(
             ['djvused', '-s', '-f', sed_file_name, djvu_path],
-            env={},  # preserve locale settings
         )
         djvused.wait()
 
@@ -119,100 +117,6 @@ class DryRunSaver(Saver):
     def save(self, document, pages, djvu_path, sed_file):
         pass
 
-class OcrEngine(object):
-
-    pass
-
-class Ocropus(OcrEngine):
-
-    name = 'ocropus'
-    has_charboxes = False
-    script_name = None
-
-    def __init__(self):
-        # Determine:
-        # - if OCRopus is installed and
-        # - which version we are dealing with
-        for script_name in 'recognize', 'rec-tess':
-            try:
-                ocropus = ipc.Subprocess(['ocroscript', script_name],
-                    stdout=ipc.PIPE,
-                    stderr=ipc.PIPE,
-                    env=dict(LC_ALL=None),  # locale=POSIX
-                )
-            except OSError:
-                raise errors.EngineNotFound(self.name)
-            try:
-                found = ocropus.stdout.read(7) == 'Usage: '
-            finally:
-                try:
-                    ocropus.wait()
-                except ipc.CalledProcessError:
-                    pass
-            if found:
-                self.script_name = script_name
-                break
-        else:
-            raise errors.EngineNotFound(self.name)
-        if script_name == 'recognize':
-            # OCRopus ≥ 0.3
-            self.has_charboxes = True
-
-    @staticmethod
-    def get_default_language():
-        return os.getenv('tesslanguage') or 'eng'
-
-    @staticmethod
-    def has_language(language):
-        return tesseract.has_language(language)
-
-    @staticmethod
-    def list_languages():
-        for language in tesseract.get_languages():
-            yield language
-
-    @contextlib.contextmanager
-    def recognize(self, pbm_file, language, details=hocr.TEXT_DETAILS_WORD):
-        def get_command_line():
-            yield 'ocroscript'
-            yield self.script_name
-            if self.has_charboxes and details < hocr.TEXT_DETAILS_LINE:
-                yield '--charboxes'
-            yield pbm_file.name
-        ocropus = ipc.Subprocess(list(get_command_line()),
-            stdout=ipc.PIPE,
-            env=dict(tesslanguage=language, LC_ALL=None),  # locale=POSIX
-        )
-        try:
-            yield ocropus.stdout
-        finally:
-            ocropus.wait()
-
-class Cuneiform(OcrEngine):
-
-    name = 'cuneiform'
-
-    def __init__(self):
-        try:
-            cuneiform.get_languages()
-        except errors.UnknownLanguageList:
-            raise errors.EngineNotFound(self.name)
-
-    @staticmethod
-    def get_default_language():
-        return 'eng'
-
-    @staticmethod
-    def has_language(language):
-        return cuneiform.has_language(language)
-
-    @staticmethod
-    def list_languages():
-        return iter(cuneiform.get_languages())
-
-    def recognize(self, pbm_file, language, details=hocr.TEXT_DETAILS_WORD):
-        return cuneiform.recognize(pbm_file, language)
-
 def get_cpu_count():
     try:
         import multiprocessing
@@ -227,8 +131,8 @@ def get_cpu_count():
 class ArgumentParser(argparse.ArgumentParser):
 
     savers = BundledSaver, IndirectSaver, ScriptSaver, InPlaceSaver, DryRunSaver
-    engines = Ocropus, Cuneiform
-    default_engine = Ocropus
+    engines = list(engines.get_engines())
+    [default_engine] = [engine for engine in engines if engine.name == 'ocropus']
 
     _details_map = dict(
         lines=hocr.TEXT_DETAILS_LINE,
@@ -376,6 +280,8 @@ class Context(djvu.decode.Context):
             self._pixel_format = bitonal_pixel_format
         else:
             self._pixel_format = rgb_pixel_format
+        if self._options.engine.image_format == 'bmp':
+            self._pixel_format.rows_top_to_bottom = 0
 
     def _temp_file(self, name, auto_remove=True):
         path = os.path.join(self._temp_dir, name)
@@ -389,35 +295,80 @@ class Context(djvu.decode.Context):
             print >>sys.stderr, message
             sys.exit(1)
 
+    @contextlib.contextmanager
+    def get_output_image(self, nth, page_job):
+        size = page_job.size
+        rect = (0, 0) + size
+        output_format = self._engine.image_format
+        file = self._temp_file('%06d.%s' % (nth, output_format))
+        try:
+            if output_format == 'ppm':
+                if self._pixel_format.bpp == 1:
+                    file.write('P4 %d %d\n' % size)  # PBM header
+                else:
+                    file.write('P6 %d %d 255\n' % size)  # PPM header
+                data = page_job.render(
+                    self._options.render_layers,
+                    rect, rect,
+                    self._pixel_format
+                )
+                file.write(data)
+                file.flush()
+            elif output_format == 'bmp':
+                dpm = int(page_job.dpi * 39.37 + 0.5)
+                data = page_job.render(
+                    self._options.render_layers,
+                    rect, rect,
+                    self._pixel_format,
+                    row_alignment=4,
+                )
+                n_palette_colors = 2 * (self._pixel_format.bpp == 1)
+                headers_size = 54 + 4 * n_palette_colors
+                file.write(struct.pack('<ccIHHI',
+                    'B', 'M', # magic
+                    len(data) + headers_size, # whole file size
+                    0, 0, # identification magic
+                    headers_size # offset to pixel data
+                ))
+                file.write(struct.pack('<IIIHHIIIIII',
+                    40, # size of this header
+                    size[0], size[1], # image size in pixels
+                    1, # number of color planes
+                    self._pixel_format.bpp, # number of bits per pixel
+                    0, # compression method
+                    len(data), # size of pixel data
+                    dpm, dpm, # resolution in pixels/meter
+                    n_palette_colors, # number of colors in the color pallete
+                    n_palette_colors # number of important colors
+                ))
+                if self._pixel_format.bpp == 1:
+                    # palette:
+                    file.write(struct.pack('<BBBB', 0, 0, 0, 0))
+                    file.write(struct.pack('<BBBB', 0xff, 0xff, 0xff, 0))
+                file.write(data)
+                file.flush()
+            else:
+                raise NotImplementedError('Unknown image format: %s' % output_format)
+            yield file
+        finally:
+            file.close()
+
     def process_page(self, page):
         print >>sys.stderr, '- Page #%d' % (page.n + 1)
         page_job = page.decode(wait=True)
         size = page_job.size
-        rect = (0, 0) + size
-        pfile = self._temp_file('%06d.pnm' % page.n)
-        try:
-            if self._pixel_format.bpp == 1:
-                pfile.write('P4 %d %d\n' % size)  # PBM header
-            else:
-                pfile.write('P6 %d %d 255\n' % size)  # PPM header
-            data = page_job.render(
-                self._options.render_layers,
-                rect, rect,
-                self._pixel_format
-            )
-            pfile.write(data)
-            pfile.flush()
-            html_file = None
+        with self.get_output_image(page.n, page_job) as pfile:
+            result_file = None
             with self._engine.recognize(pfile, language=self._options.language, details=self._options.details) as result:
                 try:
                     if self._debug:
-                        html_file = self._temp_file('%06d.html' % page.n)
-                        html = result.read()
-                        html_file.write(html)
-                        html_file.seek(0)
+                        result_file = self._temp_file('%06d.%s' % (page.n, self._engine.output_format))
+                        result_data = result.read()
+                        result_file.write(result_data)
+                        result_file.seek(0)
                     else:
-                        html_file = result
-                    text, = hocr.extract_text(html_file,
+                        result_file = result
+                    text, = self._engine.extract_text(result_file,
                         rotation=page.rotation,
                         details=self._options.details,
                         uax29=self._options.uax29,
@@ -425,10 +376,8 @@ class Context(djvu.decode.Context):
                     )
                     return text
                 finally:
-                    if html_file is not None:
-                        html_file.close()
-        finally:
-            pfile.close()
+                    if result_file is not None:
+                        result_file.close()
 
     def page_thread(self, pages, results, condition):
         for page in pages:
