@@ -19,6 +19,11 @@ PIPE = subprocess.PIPE
 
 from subprocess import CalledProcessError
 
+# Protect from scanadf[0] and possibly other brain-dead software that set
+# SIGCHLD to SIG_IGN.
+# [0] http://bugs.debian.org/596232
+signal.signal(signal.SIGCHLD, signal.SIG_DFL)
+
 def get_signal_names():
     _signame_pattern = re.compile('^SIG[A-Z0-9]*$')
     data = dict(
@@ -52,27 +57,33 @@ del get_signal_names
 
 class Subprocess(subprocess.Popen):
 
-    def __init__(self, *args, **kwargs):
-        env_override = kwargs.pop('env', {})
-        if env_override.get('LC_ALL', '') is None:
-            # Reset all locale variables
-            env = dict(
+    @classmethod
+    def override_env(cls, override):
+        env = os.environ
+        # We'd like to:
+        # - preserve LC_CTYPE (which is required by some DjVuLibre tools),
+        # - but reset all other locale settings (which tend to break things).
+        lc_ctype = env.get('LC_ALL') or env.get('LC_CTYPE') or env.get('LANG')
+        env = dict(
                 (k, v)
-                for k, v in os.environ.iteritems()
+                for k, v in env.iteritems()
                 if not (k.startswith('LC_') or k in ('LANG', 'LANGUAGES'))
-            )
-            del env_override['LC_ALL']
-        else:
-            env = dict(os.environ)
-        env.update(env_override)
-        kwargs['env'] = env
-        subprocess.Popen.__init__(self, *args, **kwargs)
+        )
+        if lc_ctype:
+            env['LC_CTYPE'] = lc_ctype
+        if override:
+            env.update(override)
+        return env
+
+    def __init__(self, *args, **kwargs):
+        kwargs['env'] = self.override_env(kwargs.get('env'))
         if os.name == 'posix':
             kwargs.update(close_fds=True)
         try:
             self.__command = kwargs['args'][0]
         except KeyError:
             self.__command = args[0][0]
+        subprocess.Popen.__init__(self, *args, **kwargs)
 
     def wait(self):
         return_code = subprocess.Popen.wait(self)
